import sys
import numpy as np
from zacrostools.analysis_functions import get_data_general, get_data_specnum


class KMCOutput:
    """A class that represents a KMC output."""

    def __init__(self, path, ignore=0.0, coverage_per_site=False, ads_sites=None):
        self.path = path

        # Get data from general_output.txt file
        data_general = get_data_general(path)
        self.n_surf_species = data_general['n_surf_species']
        self.n_sites = data_general['n_sites']
        self.area = data_general['area']
        self.site_types = data_general['site_types']

        # Get data from specnum_output.txt
        data_specnum, header = get_data_specnum(path, ignore)
        self.final_time = data_specnum[-1, 2]
        self.final_energy = data_specnum[-1, 4] / self.area
        self.energy = np.average(data_specnum[:, 4]) / self.area

        # Production (molec) and TOF (molec/s-1/Å2)
        self.production = {}  # needed for selectivity
        self.tof = {}
        for i in range(5 + self.n_surf_species, len(header)):
            ads = header[i]
            self.production[ads] = data_specnum[-1, i] - data_specnum[0, i]
            if data_specnum[-1, i] != 0:
                self.tof[header[i]] = np.polyfit(data_specnum[:, 2], data_specnum[:, i], 1)[0] / self.area
            else:
                self.tof[header[i]] = 0.00

        # Coverage (%)
        self.coverage = {}
        for i in range(5, 5 + self.n_surf_species):
            ads = header[i].replace('*', '')
            self.coverage[ads] = np.average(data_specnum[:, i]) / self.n_sites * 100
        self.total_coverage = sum(self.coverage.values())
        self.dominant_ads = max(self.coverage, key=self.coverage.get)

        # Coverage per site type (%)
        if len(self.site_types) == 1:
            print("ERROR: coverage_per_site not available when there is only one site type")
            sys.exit(f"path: {self.path}")
        if coverage_per_site:
            self.coverage_per_site_type = {}
            for site_type in self.site_types:
                self.coverage_per_site_type[site_type] = {}
            for i in range(5, 5 + self.n_surf_species):
                ads = header[i].replace('*', '')
                site_type = ads_sites[ads]
                self.coverage_per_site_type[site_type][ads] = np.average(data_specnum[:, i]) / self.site_types[
                    ads_sites[ads]] * 100
            self.total_coverage_per_site_type = {}
            self.dominant_ads_per_site_type = {}
            for site_type in self.site_types:
                self.total_coverage_per_site_type[site_type] = sum(self.coverage_per_site_type[site_type].values())
                self.dominant_ads_per_site_type[site_type] = max(self.coverage_per_site_type[site_type],
                                                                 key=self.coverage_per_site_type[site_type].get)

    def get_selectivity(self, main_product, side_products):
        selectivity = float('NaN')
        tof_side_products = 0.0
        for side_product in side_products:
            tof_side_products += self.tof[side_product]
        if self.tof[main_product] + tof_side_products != 0:
            selectivity = self.tof[main_product] / (self.tof[main_product] + tof_side_products) * 100
        return selectivity



from zacrostools.input_functions import write_header


class LatticeModel:
    """A class that represents a lattice model.

    Parameters
    ----------
    lines: list of str
        Lines that will be printed in the lattice_input.dat.
    """

    def __init__(self, lines=None):
        if isinstance(lines, list):
            self.lines = lines
        else:
            print("Error: parameter 'lines' in LatticeModel is not a list")

    @classmethod
    def from_file(cls, path):
        """Create a LatticeModel by reading an existing lattice_input.dat file.

        Parameters
        ----------
        path: str
            Path to the directorey where the lattice_input.dat file is located


        Returns
        -------
        lattice_model: LatticeModel

        """
        with open(path, 'r') as infile:
            lines = infile.readlines()
        lattice_model = cls(lines=lines)
        return lattice_model

    def write_lattice_input(self, path):
        """Write the lattice_input.dat file.

        Parameters
        ----------
        path: str
            Path to the directory where the lattice_input.dat file will be written.

        """
        write_header(f"{path}/lattice_input.dat")
        with open(f"{path}/lattice_input.dat", 'a') as infile:
            for line in self.lines:
                infile.write(line)

    def repeat_cell(self, repeat_cell):
        """Modify the value of the repeat_cell keyword in the lattice_input.dat file.

        Parameters
        ----------
        repeat_cell: list of int, optional
            Updates the repeat_cell keyword in lattice_input.dat file.

        """
        for i, line in enumerate(self.lines):
            if 'repeat_cell' in line:
                self.lines[i] = f'   repeat_cell {repeat_cell[0]} {repeat_cell[1]}\n'



import ast
import pandas as pd
from zacrostools.input_functions import write_header


class EnergeticModel:
    """A class that represents a KMC energetic model.

    Parameters:

    energetics_data: Pandas DataFrame
        Informaton on the energetic model.
        The cluster name is taken as the index of each row.

        The following columns are required:
            - cluster_eng (float): cluster formation energy (in eV)
            - site_types (str): the types of each site in the pattern
            - lattice_state (list): cluster configuration in Zacros format, e.g. ['1 CO* 1','2 CO* 1']
        The following columns are optional:
            - neighboring (str): connectivity between sites involved, e.g. 1-2. Default value: None
            - angles (str): Angle between sites in Zacros format, e.g. '1-2-3:180'. Default value: None
            - graph_multiplicity (int): symmetry number of the cluster, e.g. 2. Default value: 1
    """

    def __init__(self, energetics_data: pd.DataFrame = None):
        if isinstance(energetics_data, pd.DataFrame):
            self.df = energetics_data
        else:
            print("Error: parameter 'energetics_data' in EnergeticModel is not a Pandas DataFrame")

    @classmethod
    def from_dictionary(cls, path):
        """Not iomplemented yet"""
        pass

    def write_energetics_input(self, path):
        """Write the energetics_input.dat file"""
        write_header(f"{path}/energetics_input.dat")
        with open(f"{path}/energetics_input.dat", 'a') as infile:
            infile.write('energetics\n\n')
            infile.write('############################################################################\n\n')
            for cluster in [x for x in self.df.index if '_gas' not in x]:
                infile.write(f"cluster {cluster}\n\n")
                lattice_state = ast.literal_eval(self.df.loc[cluster, 'lattice_state'])
                infile.write(f"  sites {len(lattice_state)}\n")
                if not pd.isnull(self.df.loc[cluster, 'neighboring']):
                    infile.write(f"  neighboring {self.df.loc[cluster, 'neighboring']}\n")
                infile.write(f"  lattice_state\n")
                for element in lattice_state:
                    infile.write(f"    {element}\n")
                infile.write(f"  site_types {self.df.loc[cluster, 'site_types']}\n")
                try:
                    if not pd.isnull(self.df.loc[cluster, 'graph_multiplicity']):
                        infile.write(f"  graph_multiplicity {int(self.df.loc[cluster, 'graph_multiplicity'])}\n")
                except KeyError:
                    pass
                try:
                    if not pd.isnull(self.df.loc[cluster, 'angles']):
                        infile.write(f"  angles {self.df.loc[cluster, 'angles']}\n")
                except KeyError:
                    pass
                infile.write(f"  cluster_eng {self.df.loc[cluster, 'cluster_eng']:.2f}\n\n")
                infile.write(f"end_cluster\n\n")
                infile.write('############################################################################\n\n')
            infile.write(f"end_energetics\n")



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='zacrostools',
    version='0.4-beta',
    description='A collective of tools for the preparation of input files for ZACROS',
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    url='https://github.com/hprats/ZacrosTools',
    download_url='https://github.com/hprats/ZacrosTools/archive/refs/tags/v0.4-beta.tar.gz',
    author='Hector Prats',
    author_email='hpratsgarcia@gmail.com',
    keywords=['python', 'chemistry', 'KMC', 'ZACROS'],
    install_requires=['pandas', 'scipy']
)



import numpy as np


def find_nearest(array, value):
    """Finds the element of an array whose value is closest to a given value, and returns its index.

    Parameters
    ----------
    array: np.Array
    value: float

    Returns
    -------
    array[idx]: int

    """
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return array[idx]


def get_data_general(path):
    dmatch = ['Number of surface species',
              'Total number of lattice sites',
              'Lattice surface area',
              'Site type names and total number of sites of that type']
    data = {}
    with open(f"{path}/general_output.txt", 'r') as file_object:
        line = file_object.readline()
        while len(dmatch) != 0:
            if 'Number of surface species' in line:
                data['n_surf_species'] = int(line.split()[-1])
                dmatch.remove('Number of surface species')
            if 'Total number of lattice sites' in line:
                data['n_sites'] = int(line.split()[-1])
                dmatch.remove('Total number of lattice sites')
            if 'Lattice surface area' in line:
                data['area'] = float(line.split()[-1])
                dmatch.remove('Lattice surface area')
            if 'Site type names and total number of sites of that type' in line:
                line = file_object.readline()
                site_types = {}
                while line.strip():
                    num_sites_of_given_type = int(line.strip().split(' ')[1].replace('(', '').replace(')', ''))
                    site_types[line.strip().split(' ')[0]] = num_sites_of_given_type
                    line = file_object.readline()
                data['site_types'] = site_types
                dmatch.remove('Site type names and total number of sites of that type')
            line = file_object.readline()
        return data


def get_data_specnum(path, ignore=0.0):
    with open(f"{path}/specnum_output.txt", "r") as infile:
        header = infile.readline().split()
    full_data = np.loadtxt(f"{path}/specnum_output.txt", skiprows=1)
    index = np.where(full_data[:, 2] == find_nearest(full_data[:, 2], full_data[-1, 2] * ignore / 100))[0][0]
    data = np.delete(full_data, slice(0, index), 0)
    return data, header


def get_data_simulation(path):
    dmatch = ['pressure',
              'gas_specs_names',
              'gas_molar_fracs']
    data = {}
    with open(f"{path}/simulation_input.dat", 'r') as file_object:
        line = file_object.readline()
        while len(dmatch) != 0:
            if 'pressure' in line:
                data['pressure'] = float(line.split()[-1])
                dmatch.remove('pressure')
            if 'gas_specs_names' in line:
                data['gas_specs_names'] = line.split()[1:]
                dmatch.remove('gas_specs_names')
            if 'gas_molar_fracs' in line:
                data['gas_molar_fracs'] = [float(x) for x in line.split()[1:]]
                dmatch.remove('gas_molar_fracs')
            line = file_object.readline()
        return data
